#ifndef AVIFILE_STREAMINFO_H
#define AVIFILE_STREAMINFO_H

// this file could be a part of infotypes in future  > so be careful


#include "default.h"
//#include <ostream.h>

class IMediaReadStream;
class AviMediaReadStream;
class ASFReadStream;

class StreamInfo
{
    friend class IMediaReadStream;
    friend class AviMediaReadStream;
    friend class ASFReadStream;
public:
    enum Type { Video, Audio, Other };

    StreamInfo();
    ~StreamInfo();

    Type GetType() const;
    fourcc_t GetFourCC() const;

    double GetBps() const;	// calculated Bps
    double GetFps() const;	// calculated Fps

    double GetLengthTime() const;
    int64_t GetStreamSize() const;
    uint_t GetStreamFrames() const;

    uint_t GetMaxKfFrameSize() const;
    uint_t GetMinKfFrameSize() const;
    uint_t GetAvgKfFrameSize() const;
    uint_t GetKfFrames() const;
    int64_t GetKfFramesSize() const;

    uint_t GetMaxFrameSize() const;
    uint_t GetMinFrameSize() const;
    uint_t GetAvgFrameSize() const;
    uint_t GetFrames() const;
    int64_t GetFramesSize() const;

    // for non video returns -1
    int GetVideoWidth() const;
    int GetVideoHeight() const;

    // for non audio returns -1
    int GetAudioChannels() const;
    int GetAudioSamplesPerSec() const;
    int GetAudioBitsPerSample() const;
    bool IsAudioVbr() const;//??? no sure
    avm::string GetString() const;
//    friend ostream & operator << (ostream&, StreamInfo &);
//    friend ostream & operator << (ostream&, StreamInfo *);
protected:
    Type m_Type;
    fourcc_t m_FourCC;		// fourcc type of stream

    double m_dLengthTime;	// time in seconds for all packets
    uint_t m_uiMaxKfFrameSize;	// biggest keyframe chunk size
    uint_t m_uiMinKfFrameSize;	// smallest keyframe chunk size
    uint_t m_uiKfFrames;	// keyframe chunks
    int64_t m_uiKfFramesSize;	// total keyframe chunk size

    // should make sence probably only for video stream
    uint_t m_uiMaxFrameSize;	// biggest non-keyframe chunk size
    uint_t m_uiMinFrameSize;	// smallest non-keyframe chunk size
    uint_t m_uiFrames;		// non-keyframe chunks
    int64_t m_uiFramesSize;	// total non-keyframe chunk size

    union
    {
        char b[512];
	struct
	{
	    int	m_iChannels;
	    int	m_iSamplesPerSec;
	    int	m_iBitsPerSample;
            bool m_bVbr;
	} aud;
	struct
	{
	    int m_iWidth;
	    int m_iHeight;

	} vid;
    } u;
};

#endif // AVIFILE_STREAMSTATS_H
