#include "StreamInfo.h"
#include <stdio.h>

using namespace avm;

StreamInfo::StreamInfo()
{
    m_dLengthTime = 0.0;
    m_uiMaxKfFrameSize = 0;
    m_uiMinKfFrameSize = 0;
    m_uiKfFrames = 0;
    m_uiKfFramesSize = 0;

    m_uiMaxFrameSize = 0;
    m_uiMinFrameSize = 0;
    m_uiFrames = 0;
    m_uiFramesSize = 0;

    m_Type = Video;
}

StreamInfo::~StreamInfo()
{
}

StreamInfo::Type StreamInfo::GetType() const
{
    return m_Type;
}

fourcc_t StreamInfo::GetFourCC() const
{
    return m_FourCC;
}

double StreamInfo::GetBps() const
{
    return (m_dLengthTime) ? GetStreamSize() / m_dLengthTime : 0.0;
}

double StreamInfo::GetFps() const
{
    return (m_dLengthTime) ? GetStreamFrames() /  m_dLengthTime : 0.0;
}

double StreamInfo::GetLengthTime() const
{
    return m_dLengthTime;
}

int64_t StreamInfo::GetStreamSize() const
{
    return m_uiKfFramesSize + m_uiFramesSize;
}

uint_t StreamInfo::GetStreamFrames() const
{
    return m_uiKfFrames + m_uiFrames;
}

uint_t  StreamInfo::GetMaxKfFrameSize() const
{
    return m_uiMaxKfFrameSize;
}

uint_t  StreamInfo::GetMinKfFrameSize() const
{
    return m_uiMinKfFrameSize;
}

uint_t  StreamInfo::GetAvgKfFrameSize() const
{
    return (m_uiKfFrames > 0) ? m_uiKfFramesSize / m_uiKfFrames : 0;
}

int64_t StreamInfo::GetKfFramesSize() const
{
    return m_uiKfFramesSize;
}

uint_t StreamInfo::GetKfFrames() const
{
    return m_uiKfFrames;
}

uint_t  StreamInfo::GetMaxFrameSize() const
{
    return m_uiMaxFrameSize;
}

uint_t  StreamInfo::GetMinFrameSize() const
{
    return m_uiMinFrameSize;
}

uint_t  StreamInfo::GetAvgFrameSize() const
{
    return (m_uiFrames > 0) ? m_uiFramesSize / m_uiFrames : 0;
}

int64_t StreamInfo::GetFramesSize() const
{
    return m_uiFramesSize;
}

uint_t StreamInfo::GetFrames() const
{
    return m_uiFrames;
}

int StreamInfo::GetAudioChannels() const
{
    return (m_Type == Audio) ? u.aud.m_iChannels : -1;
}

int StreamInfo::GetAudioSamplesPerSec() const
{
    return (m_Type == Audio) ? u.aud.m_iSamplesPerSec : -1;
}

int StreamInfo::GetAudioBitsPerSample() const
{
    return (m_Type == Audio) ? u.aud.m_iBitsPerSample : -1;
}

bool StreamInfo::IsAudioVbr() const
{
    return (m_Type == Audio) ? u.aud.m_bVbr : false;
}

string StreamInfo::GetString() const
{
    double keyproc = 0, delproc = 0;
    double skeyproc = 0, sdelproc = 0;
    char buffer[2048];
    if (GetStreamFrames() > 0)
    {
	keyproc = GetKfFrames()/(double)GetStreamFrames() * 100.0;
	delproc = 100.0 - keyproc;
	skeyproc = (GetStreamSize() > 0) ? GetKfFramesSize()/(double)GetStreamSize() : 0.0;
	skeyproc *= 100.0;
	sdelproc = 100.0 - skeyproc;
    }
    sprintf(buffer, "  Time length: %f  Fps: %f  Size: %Ld bytes (%.2fKB)\n"
	    "  Frames: %d    Bps: %.2f kbps (%.2fKB/s)\n"
	    "  Key frames/size %.2f%% (%.2f%%)  delta frames/size %.2f%%\n"
	    "  Min/avg/max/total key frame size: %d/%d/%d (%.2fKB) in %d frames\n"
	    "  Min/avg/max/total delta frame size: %d/%d/%d (%.2fKB) in %d frames\n",
	    GetLengthTime(),
	    GetFps(),
	    GetStreamSize(),
	    GetStreamSize() / 1024.0,

	    GetStreamFrames(),
	    GetBps() * 8 / 1000,
	    GetBps() / 1024.0,

	    keyproc, skeyproc, sdelproc,

	    GetMinKfFrameSize(),
	    GetAvgKfFrameSize(),
	    GetMaxKfFrameSize(),
	    GetKfFramesSize() / 1024.0,
	    GetKfFrames(),

	    GetMinFrameSize(),
	    GetAvgFrameSize(),
	    GetMaxFrameSize(),
	    GetFramesSize() / 1024.0,
	    GetFrames());

    return string(buffer);
}
