#include "MemQueue.h"

#include <stdio.h>

using namespace std;
/*
 * Smart data queue
 *
 * should limit memory copying to the minimum
 */


MemQueue::MemQueue(uint_t bufsize, uint_t maxfrag)
{
    maxfragment = maxfrag;
    alloc = bufsize + 2 * maxfragment;
    buffer = new uint8_t[alloc];
    length = inpos = outpos = 0;
    lastpos = 0;
    //lastposcpy = 0;
    maxpos = bufsize;
}

MemQueue::~MemQueue()
{
    delete buffer;
}

void MemQueue::Add(uint_t bytes)
{
    Locker locker(m_Mutex);
    if ((length + bytes) < maxpos)
    {
	inpos += bytes;
	if (inpos > (int)maxpos)
	{
	    lastpos = inpos;
	    inpos = 0;
	}
	length += bytes;
    }
    else printf("MemQueue::Add(%d) ERROR!\n", bytes);

}

void MemQueue::Free(uint_t bytes)
{
    Locker locker(m_Mutex);
    if (length >= bytes)
    {
	outpos += bytes;
	if (outpos >= lastpos)
	{
	    outpos -= lastpos;
	    //outpos += lastposcpy;
	    lastpos = 0;
	}
	length -= bytes;
    }
    else printf("MemQueue::Free(%d) ERROR!\n", bytes);
}

uint8_t* MemQueue::GetFree(uint_t bytes)
{
    Locker locker(m_Mutex);

    if (maxfragment < bytes)
    {
	printf("bigger fragment found %d   (%d)\n", bytes, maxfragment);
        int na = alloc - 2 * maxfragment;
	maxfragment = 2 * bytes;
	na += 2 * maxfragment;
	uint8_t* nb = new uint8_t[na];
	memcpy(nb, buffer, alloc);
        delete buffer;
	buffer = nb;
        alloc = na;
    }

    int f = maxpos - length;
    if (f > (int)bytes)
	return buffer + inpos;
    return 0;
}

uint8_t* MemQueue::GetData(uint_t bytes)
{
    Locker locker(m_Mutex);
    if (length > bytes)
    {
        int lp = outpos + bytes;
	if (lp > lastpos)
	{
	    //if ()
            // copy from begining to make contiuos space
	    memcpy(buffer + lastpos, buffer, (outpos + bytes) - lastpos);
	    printf("Copy data: %d\n", (outpos + bytes) - lastpos);
	}
	return buffer + outpos;
    }
    return 0;
}

void MemQueue::DebugPrint()
{
    printf("Data  in:%d  out:%d  len:%d   lastpos:%d\n",
	   outpos, inpos, length, lastpos);
}



int main(int argc, char* argv[])
{
    MemQueue d(50, 5);

    char test[] = "0123456789";


    for (int i = 0; i < 100; i++)
    {
	void* b = d.GetFree(9);
	if (b)
	{
	    memcpy(b, test, 9);
	    d.Add(9);
	}
	else
	{
	    char* s = (char*) d.GetData(7);
	    printf("read 7   %.7s\n", s);
            d.Free(7);
	}
	d.DebugPrint();
    }

}
